'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _objectWithoutPropertiesLoose = _interopDefault(require('@babel/runtime/helpers/objectWithoutPropertiesLoose'));
var _extends = _interopDefault(require('@babel/runtime/helpers/extends'));
var THREE = require('three');
var React = require('react');
var React__default = _interopDefault(React);
var Reconciler = _interopDefault(require('react-reconciler'));
var omit = _interopDefault(require('lodash-es/omit'));
var upperFirst = _interopDefault(require('lodash-es/upperFirst'));
var ResizeObserver = _interopDefault(require('resize-observer-polyfill'));
var scheduler = require('scheduler');

var roots = new Map();
var emptyObject = {};
function applyProps(instance, newProps, oldProps) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  // Filter equals, events and reserved props
  var sameProps = Object.keys(newProps).filter(function (key) {
    return newProps[key] === oldProps[key];
  });
  var handlers = Object.keys(newProps).filter(function (key) {
    return typeof newProps[key] === 'function' && key.startsWith('on');
  });
  var filteredProps = omit(newProps, [].concat(sameProps, handlers, ['children', 'key', 'ref']));

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach(function (_ref) {
      var key = _ref[0],
          value = _ref[1];
      var root = instance;
      var target = root[key];

      if (key.includes('-')) {
        var entries = key.split('-');
        target = entries.reduce(function (acc, key) {
          return acc[key];
        }, instance);

        if (target && !target.set) {
          // The target is atomic, this forces us to switch the root
          var _entries$reverse = entries.reverse(),
              name = _entries$reverse[0],
              reverseEntries = _entries$reverse.slice(1);

          root = reverseEntries.reverse().reduce(function (acc, key) {
            return acc[key];
          }, instance);
          key = name;
        }
      }

      if (target && target.set) {
        if (target.constructor.name === value.constructor.name) {
          target.copy(value);
        } else if (Array.isArray(value)) {
          var _target;

          (_target = target).set.apply(_target, value);
        } else {
          target.set(value);
        }
      } else root[key] = value;
    });

    if (handlers.length) {
      instance.__handlers = handlers.reduce(function (acc, key) {
        var _extends2;

        var name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, (_extends2 = {}, _extends2[name] = newProps[key], _extends2));
      }, {});
    }
  }
}
var Renderer = Reconciler({
  supportsMutation: true,
  isPrimaryRenderer: false,
  now: scheduler.unstable_now,
  getPublicInstance: function getPublicInstance(instance) {
    return instance;
  },
  getRootHostContext: function getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },
  getChildHostContext: function getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },
  createInstance: function createInstance(type, props, rootContainerInstance, hostContext, internalInstanceHandle) {
    var instance = type === 'primitive' ? props.object : new THREE[upperFirst(type)]();
    applyProps(instance, props, {});
    return instance;
  },
  createTextInstance: function createTextInstance() {},
  appendInitialChild: function appendInitialChild(parentInstance, child) {
    if (child) parentInstance.add(child);
  },
  finalizeInitialChildren: function finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },
  prepareUpdate: function prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },
  shouldDeprioritizeSubtree: function shouldDeprioritizeSubtree(type, props) {
    return false;
  },
  prepareForCommit: function prepareForCommit() {},
  resetAfterCommit: function resetAfterCommit() {},
  shouldSetTextContent: function shouldSetTextContent(props) {
    return false;
  },
  appendChild: function appendChild(parentInstance, child) {
    if (child) parentInstance.add(child);
  },
  appendChildToContainer: function appendChildToContainer(parentInstance, child) {
    if (child) parentInstance.add(child);
  },
  insertBefore: function insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      child.parent = parentInstance;
      child.dispatchEvent({
        type: 'added'
      });
      var index = parentInstance.children.indexOf(beforeChild);
      parentInstance.children = [].concat(parentInstance.children.slice(0, index), [child], parentInstance.children.slice(index));
    }
  },
  removeChild: function removeChild(parentInstance, child) {
    if (child) parentInstance.remove(child);
  },
  removeChildFromContainer: function removeChildFromContainer(parentInstance, child) {
    if (child) parentInstance.remove(child);
  },
  commitUpdate: function commitUpdate(instance, updatePayload, type, oldProps, newProps) {
    instance.busy = true;
    applyProps(instance, newProps, oldProps);
    instance.busy = false;
  },
  schedulePassiveEffects: scheduler.unstable_scheduleCallback,
  cancelPassiveEffects: scheduler.unstable_cancelCallback
});
function render(element, container) {
  var root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  var root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, function () {
    return roots.delete(container);
  });
}
function useMeasure() {
  var ref = React.useRef();

  var _useState = React.useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
      bounds = _useState[0],
      set = _useState[1];

  var _useState2 = React.useState(function () {
    return new ResizeObserver(function (_ref2) {
      var entry = _ref2[0];
      return set(entry.contentRect);
    });
  }),
      ro = _useState2[0];

  React.useEffect(function () {
    if (ref.current) ro.observe(ref.current);
    return function () {
      return ro.disconnect();
    };
  }, []);
  return [{
    ref: ref
  }, bounds];
}
var context = React__default.createContext();
function Canvas(_ref3) {
  var children = _ref3.children,
      style = _ref3.style,
      camera = _ref3.camera,
      onCreated = _ref3.onCreated,
      onUpdate = _ref3.onUpdate,
      props = _objectWithoutPropertiesLoose(_ref3, ["children", "style", "camera", "onCreated", "onUpdate"]);

  var canvasRef = React.useRef();
  var renderer = React.useRef();
  var cameraRef = React.useRef();
  var scene = React.useRef();
  var active = React.useRef(true);

  var _useMeasure = useMeasure(),
      bind = _useMeasure[0],
      measurements = _useMeasure[1];

  var _useState3 = React.useState(function () {
    return new THREE.Raycaster();
  }),
      raycaster = _useState3[0];

  var _useState4 = React.useState(function () {
    return new THREE.Vector2();
  }),
      mouse = _useState4[0];

  var subscribers = React.useRef([]);

  var _useState5 = React.useState('default'),
      cursor = _useState5[0],
      setCursor = _useState5[1];

  React.useEffect(function () {
    scene.current = new THREE.Scene();
    renderer.current = new THREE.WebGLRenderer({
      canvas: canvasRef.current,
      antialias: true
    });
    renderer.current.setClearColor(new THREE.Color('transparent'));
    cameraRef.current = camera && camera.current || new THREE.PerspectiveCamera(75, 0, 0.1, 1000);
    renderer.current.setSize(0, 0, false);
    cameraRef.current.position.z = 5;
    if (onCreated) onCreated(renderer.current, scene.current, cameraRef.current);

    var renderLoop = function renderLoop() {
      if (!active.current) return;
      requestAnimationFrame(renderLoop);
      if (onUpdate) onUpdate();
      subscribers.current.forEach(function (fn) {
        return fn();
      });
      renderer.current.render(scene.current, cameraRef.current);
    };

    render(React__default.createElement(context.Provider, {
      value: {
        scene: scene.current,
        renderer: renderer.current,
        camera: cameraRef.current,
        subscribe: function subscribe(fn) {
          subscribers.current.push(fn);
          return function () {
            return subscribers.current = subscribers.current.filter(function (s) {
              return s === fn;
            });
          };
        }
      }
    }, children), scene.current);
    requestAnimationFrame(renderLoop);
    return function () {
      active.current = false;
      unmountComponentAtNode(scene.current);
    };
  }, []);
  React.useEffect(function () {
    renderer.current.setSize(measurements.width, measurements.height, false);
    var aspect = measurements.width / measurements.height;
    cameraRef.current.aspect = aspect;
    cameraRef.current.updateProjectionMatrix();
    cameraRef.current.radius = (measurements.width + measurements.height) / 4;
  });
  var intersect = React.useCallback(function (event, fn) {
    mouse.x = event.clientX / measurements.width * 2 - 1;
    mouse.y = -(event.clientY / measurements.height) * 2 + 1;
    raycaster.setFromCamera(mouse, cameraRef.current);
    var intersects = raycaster.intersectObjects(scene.current.children, true);

    for (var i = 0; i < intersects.length; i++) {
      if (!intersects[i].object.__handlers) continue;
      fn(intersects[i]);
    }

    return intersects;
  });
  React.useEffect(function () {
    var hovered = {};

    var handleMove = function handleMove(event) {
      var hover = false;
      var intersects = intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(function (object) {
        if (!intersects.length || !intersects.find(function (i) {
          return i.object === object;
        })) {
          if (object.__handlers.unhover) object.__handlers.unhover();
          delete hovered[object.uuid];
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    return function () {
      window.removeEventListener('mousemove', handleMove);
    };
  });
  return React__default.createElement("div", _extends({}, bind, props, {
    onClick: function onClick(event) {
      var intersects = intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;
        if (handlers.click) handlers.click(data);
      });
    },
    style: _extends({
      cursor: cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }, style)
  }), React__default.createElement("canvas", {
    ref: canvasRef
  }));
}

exports.applyProps = applyProps;
exports.render = render;
exports.unmountComponentAtNode = unmountComponentAtNode;
exports.useMeasure = useMeasure;
exports.context = context;
exports.Canvas = Canvas;
