import _objectWithoutPropertiesLoose from '@babel/runtime/helpers/esm/objectWithoutPropertiesLoose';
import _extends from '@babel/runtime/helpers/esm/extends';
import * as THREE from 'three';
import { Raycaster, Vector2, Scene, WebGLRenderer, Color, PerspectiveCamera } from 'three';
import React, { useRef, useState, useEffect, useCallback } from 'react';
import Reconciler from 'react-reconciler';
import omit from 'lodash-es/omit';
import upperFirst from 'lodash-es/upperFirst';
import ResizeObserver from 'resize-observer-polyfill';
import { unstable_now, unstable_scheduleCallback, unstable_cancelCallback } from 'scheduler';

const roots = new Map();
const emptyObject = {};
function applyProps(instance, newProps, oldProps) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  // Filter equals, events and reserved props
  const sameProps = Object.keys(newProps).filter(key => newProps[key] === oldProps[key]);
  const handlers = Object.keys(newProps).filter(key => typeof newProps[key] === 'function' && key.startsWith('on'));
  const filteredProps = omit(newProps, [...sameProps, ...handlers, 'children', 'key', 'ref']);

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach((_ref) => {
      let key = _ref[0],
          value = _ref[1];
      let root = instance;
      let target = root[key];

      if (key.includes('-')) {
        const entries = key.split('-');
        target = entries.reduce((acc, key) => acc[key], instance);

        if (target && !target.set) {
          // The target is atomic, this forces us to switch the root
          const _entries$reverse = entries.reverse(),
                name = _entries$reverse[0],
                reverseEntries = _entries$reverse.slice(1);

          root = reverseEntries.reverse().reduce((acc, key) => acc[key], instance);
          key = name;
        }
      }

      if (target && target.set) {
        if (target.constructor.name === value.constructor.name) {
          target.copy(value);
        } else if (Array.isArray(value)) {
          target.set(...value);
        } else {
          target.set(value);
        }
      } else root[key] = value;
    });

    if (handlers.length) {
      instance.__handlers = handlers.reduce((acc, key) => {
        const name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, {
          [name]: newProps[key]
        });
      }, {});
    }
  }
}
const Renderer = Reconciler({
  supportsMutation: true,
  isPrimaryRenderer: false,
  now: unstable_now,

  getPublicInstance(instance) {
    return instance;
  },

  getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },

  getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },

  createInstance(type, props, rootContainerInstance, hostContext, internalInstanceHandle) {
    const instance = type === 'primitive' ? props.object : new THREE[upperFirst(type)]();
    applyProps(instance, props, {});
    return instance;
  },

  createTextInstance() {},

  appendInitialChild(parentInstance, child) {
    if (child) parentInstance.add(child);
  },

  finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },

  prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },

  shouldDeprioritizeSubtree(type, props) {
    return false;
  },

  prepareForCommit() {},

  resetAfterCommit() {},

  shouldSetTextContent(props) {
    return false;
  },

  appendChild(parentInstance, child) {
    if (child) parentInstance.add(child);
  },

  appendChildToContainer(parentInstance, child) {
    if (child) parentInstance.add(child);
  },

  insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      child.parent = parentInstance;
      child.dispatchEvent({
        type: 'added'
      });
      const index = parentInstance.children.indexOf(beforeChild);
      parentInstance.children = [...parentInstance.children.slice(0, index), child, ...parentInstance.children.slice(index)];
    }
  },

  removeChild(parentInstance, child) {
    if (child) parentInstance.remove(child);
  },

  removeChildFromContainer(parentInstance, child) {
    if (child) parentInstance.remove(child);
  },

  commitUpdate(instance, updatePayload, type, oldProps, newProps) {
    instance.busy = true;
    applyProps(instance, newProps, oldProps);
    instance.busy = false;
  },

  schedulePassiveEffects: unstable_scheduleCallback,
  cancelPassiveEffects: unstable_cancelCallback
});
function render(element, container) {
  let root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  const root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, () => roots.delete(container));
}
function useMeasure() {
  const ref = useRef();

  const _useState = useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
        bounds = _useState[0],
        set = _useState[1];

  const _useState2 = useState(() => new ResizeObserver((_ref2) => {
    let entry = _ref2[0];
    return set(entry.contentRect);
  })),
        ro = _useState2[0];

  useEffect(() => {
    if (ref.current) ro.observe(ref.current);
    return () => ro.disconnect();
  }, []);
  return [{
    ref
  }, bounds];
}
const context = React.createContext();
function Canvas(_ref3) {
  let children = _ref3.children,
      style = _ref3.style,
      camera = _ref3.camera,
      onCreated = _ref3.onCreated,
      onUpdate = _ref3.onUpdate,
      props = _objectWithoutPropertiesLoose(_ref3, ["children", "style", "camera", "onCreated", "onUpdate"]);

  const canvasRef = useRef();
  const renderer = useRef();
  const cameraRef = useRef();
  const scene = useRef();
  const active = useRef(true);

  const _useMeasure = useMeasure(),
        bind = _useMeasure[0],
        measurements = _useMeasure[1];

  const _useState3 = useState(() => new Raycaster()),
        raycaster = _useState3[0];

  const _useState4 = useState(() => new Vector2()),
        mouse = _useState4[0];

  const subscribers = useRef([]);

  const _useState5 = useState('default'),
        cursor = _useState5[0],
        setCursor = _useState5[1];

  useEffect(() => {
    scene.current = new Scene();
    renderer.current = new WebGLRenderer({
      canvas: canvasRef.current,
      antialias: true
    });
    renderer.current.setClearColor(new Color('transparent'));
    cameraRef.current = camera && camera.current || new PerspectiveCamera(75, 0, 0.1, 1000);
    renderer.current.setSize(0, 0, false);
    cameraRef.current.position.z = 5;
    if (onCreated) onCreated(renderer.current, scene.current, cameraRef.current);

    const renderLoop = function renderLoop() {
      if (!active.current) return;
      requestAnimationFrame(renderLoop);
      if (onUpdate) onUpdate();
      subscribers.current.forEach(fn => fn());
      renderer.current.render(scene.current, cameraRef.current);
    };

    render(React.createElement(context.Provider, {
      value: {
        scene: scene.current,
        renderer: renderer.current,
        camera: cameraRef.current,
        subscribe: fn => {
          subscribers.current.push(fn);
          return () => subscribers.current = subscribers.current.filter(s => s === fn);
        }
      }
    }, children), scene.current);
    requestAnimationFrame(renderLoop);
    return () => {
      active.current = false;
      unmountComponentAtNode(scene.current);
    };
  }, []);
  useEffect(() => {
    renderer.current.setSize(measurements.width, measurements.height, false);
    const aspect = measurements.width / measurements.height;
    cameraRef.current.aspect = aspect;
    cameraRef.current.updateProjectionMatrix();
    cameraRef.current.radius = (measurements.width + measurements.height) / 4;
  });
  const intersect = useCallback((event, fn) => {
    mouse.x = event.clientX / measurements.width * 2 - 1;
    mouse.y = -(event.clientY / measurements.height) * 2 + 1;
    raycaster.setFromCamera(mouse, cameraRef.current);
    const intersects = raycaster.intersectObjects(scene.current.children, true);

    for (var i = 0; i < intersects.length; i++) {
      if (!intersects[i].object.__handlers) continue;
      fn(intersects[i]);
    }

    return intersects;
  });
  useEffect(() => {
    const hovered = {};

    const handleMove = event => {
      let hover = false;
      let intersects = intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(object => {
        if (!intersects.length || !intersects.find(i => i.object === object)) {
          if (object.__handlers.unhover) object.__handlers.unhover();
          delete hovered[object.uuid];
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    return () => {
      window.removeEventListener('mousemove', handleMove);
    };
  });
  return React.createElement("div", _extends({}, bind, props, {
    onClick: event => {
      let intersects = intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;
        if (handlers.click) handlers.click(data);
      });
    },
    style: _extends({
      cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }, style)
  }), React.createElement("canvas", {
    ref: canvasRef
  }));
}

export { applyProps, render, unmountComponentAtNode, useMeasure, context, Canvas };
